"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""
import numpy as np

from ANNarchy.parser.AnalyseNeuron import analyse_neuron
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages

class Neuron :
    """
    Base class to define a neuron model.

    Neurons are rate-coded by default (in which case they must define the variable `r`). `parameters` expects a dictionary of parameter values, `equations` expects a list of variables. 
    
    Spiking neurons must define the `spike` condition (and usually also `reset`). They do not need to define `r`. 

    ```python
    LIF = ann.Neuron(
        parameters = dict(
            tau = 10.0
        ),
        equations = [
            "tau * dv/dt + v = g_exc",
        ],
        spike = "v > 30.0",
        reset = "v = 0.0",
        refractory = 5.0,
        name = "LIF",
        description = "Leaky Integrate-and-Fire spiking neuron with time constant $\\tau$." 
    )
    ```

    :param parameters: parameters of the neuron and their initial value.
    :param equations: equations defining the temporal evolution of variables.
    :param functions: additional functions used in the variables' equations.
    :param spike: condition to emit a spike (only for spiking neurons).
    :param axon_spike: condition to emit an axonal spike (only for spiking neurons and optional). The axonal spike can appear additional to the spike and is independent from refractoriness of a neuron.
    :param reset: changes to the variables after a spike (only for spiking neurons).
    :param axon_reset: changes to the variables after an axonal spike (only for spiking neurons).
    :param refractory: refractory period of a neuron after a spike (only for spiking neurons).
    :param name: name of the neuron type (used for reporting only).
    :param description: short description of the neuron type (used for reporting).
    """

    # Default name and description for reporting
    _default_names = {'rate': "Rate-coded neuron", 'spike': "Spiking neuron"}

    def __init__(self, parameters:str|dict="", equations:str|list="", spike:str=None, axon_spike:str=None, reset:str|list=None, axon_reset:str|list=None, refractory:str = None, functions:str=None, name:str="", description:str="", extra_values:dict={} ):

        # Store the parameters and equations
        self.parameters = parameters
        self.equations = equations
        self.functions = functions
        self.spike = spike
        self.axon_spike = axon_spike
        self.reset = reset
        self.axon_reset = axon_reset
        self.refractory = refractory
        self.extra_values = extra_values

        # Find the type of the neuron
        self.type = 'spike' if self.spike else 'rate'

        # Not available by now ...
        if axon_spike and ConfigManager().get('paradigm', 0) != "openmp":
            Messages._error("Axonal spike conditions are only available for openMP by now.")
            # will crash when paradigm='cuda' is passed only at the Network level...

        # Reporting
        if not hasattr(self, '_instantiated') : # User-defined
            GlobalObjectManager().add_neuron_type(self)
        elif len(self._instantiated) == 0: # First instantiated of the class
            GlobalObjectManager().add_neuron_type(self)
        self._rk_neurons_type = GlobalObjectManager().num_neuron_types()

        if name:
            self.name = name
        else:
            self.name = self._default_names[self.type]

        if description:
            self.short_description = description
        else:
            self.short_description = "User-defined model of a spiking neuron." if self.type == 'spike' else "User-defined model of a rate-coded neuron."

        # Analyse the neuron type
        self.description = None

    def _analyse(self, net_id):
        # Analyse the neuron type
        if not self.description:
            self.description = analyse_neuron(self, net_id)

    def __repr__(self):
        if self.type == 'rate':
            text= """Rate-coded neuron.

Parameters:
""" + str(self.parameters) + """
Equations of the variables:
""" + str(self.equations) + """

"""
        else:
            text= """Spiking neuron.

Parameters:
""" + str(self.parameters) + """
Equations of the variables:
""" + str(self.equations) + """
Spiking condition:
""" + str(self.spike) + """
Axonal spiking condition:
""" + str(self.axon_spike) + """
Reset after a spike:
""" + str(self.reset)

        return text


class IndividualNeuron :
    """
    Neuron object returned by the Population.neuron(rank) method.

    This only a wrapper around the Population data. It has the same attributes (parameter and variable) as the original population.
    """
    def __init__(self, population, rank):
        self.population  = population
        self.rank  = rank

    def __getattr__(self, name):
        " Method called when accessing an attribute."
        if name == 'population':
            return object.__getattribute__(self, name)
        elif hasattr(self.population, 'attributes'):
            if name in self.population.attributes:
                if not self.population.initialized: # Store it in the temporary array
                    val = (self.population.get(name))
                    if isinstance(val, np.ndarray):
                        return val[self.population.coordinates_from_rank(self.rank)]
                    else:
                        return val
                else:
                    if _check_paradigm("cuda", self.population.net_id):
                        self.population.cyInstance.device_to_host(name)
                    if name in self.population.neuron_type.description['local']:
                        var = getattr(self.population.cyInstance, name)
                        return var[self.rank]
                    else:
                        return getattr(self.population.cyInstance, name)
            else:
                return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        " Method called when setting an attribute."
        if name == 'population':
            object.__setattr__(self, name, value)
        elif name == 'rank':
            object.__setattr__(self, name, value)
        elif hasattr(self.population, 'attributes'):
            if name in self.population.attributes:
                if name in self.population.neuron_type.description['local']:
                    if not self.population.initialized: # Store it in the temporary array
                        newval = self.population.get(name)
                        newval[self.population.coordinates_from_rank(self.rank)] = value
                        self.population.__setattr__(name, newval)
                    else: # Access the C++ data
                        var = getattr(self.population.cyInstance, name)
                        var[self.rank] = value
                        setattr(self.population.cyInstance, name, var)
                else:
                    self.population.__setattr__(name, value)
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def __repr__(self):
        desc = 'Neuron of the population ' + self.population.name + ' with rank ' + str(self.rank) + ' (coordinates ' + str(self.population.coordinates_from_rank(self.rank)) + ').\n'
        desc += 'Parameters:\n'
        for param in self.population.parameters:
            desc += '  ' + param + ' = ' + str(self.__getattr__(param)) + '\n'
        desc += '\nVariables:\n'
        for param in self.population.variables:
            desc += '  ' + param + ' = ' + str(self.__getattr__(param)) + '\n'
        return desc

    def __add__(self, other):
        """Allows to join two neurons if they have the same population."""
        if other.population == self.population:
            if isinstance(other, IndividualNeuron):
                return PopulationView(self.population, list(set([self.rank, other.rank])))
            elif isinstance(other, PopulationView):
                return PopulationView(self.population, list(set([self.rank] + other.ranks)))
        else:
            Messages._error("can only add two PopulationViews of the same population.")
            return None



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
additional_global_functions = ""

#TODO: maybe it would make more sense to split up the nb_blocks and tpb for ELL and COO
launch_config = {
    'init': """
        _threads_per_block = 64;
        _nb_blocks = std::min<unsigned int>(nb_dendrites(), 65535);

    #ifdef _DEBUG
        std::cout << "Initial kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
        if (nb_blocks != -1) {
            _nb_blocks = static_cast<unsigned int>(nb_blocks);
            _threads_per_block = threads_per_block;
        }else{
            _threads_per_block = threads_per_block;
            _nb_blocks = std::min<unsigned int>(nb_dendrites(), 65535);
        }

    #ifdef _DEBUG
        std::cout << "Updated kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    hyb_local<%(type)s>* %(name)s;
    hyb_local_gpu<%(type)s>* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': "",
    'global': {
        'parameter':
    """
    // Global parameter %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global variable %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
"""
    }
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
""",
    'semiglobal': "",
    'global': {
        'parameter': """
        // Global parameter %(name)s
        %(name)s = 0.0;
""",
        'variable': """
        // Global variable %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
    // Local %(name)s
    size_in_bytes += %(name)s->size_in_bytes();
""",
    'semiglobal': "",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s->clear();

        // %(name)s - device
        gpu_%(name)s->clear();
""",
    'semiglobal': "",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s->ell, %(name)s->ell.data(), %(name)s->ell.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            cudaMemcpy( gpu_%(name)s->coo, %(name)s->coo.data(), %(name)s->coo.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': "",
    'global': ""
}

attribute_device_to_host = {
    'local': "",
    'semiglobal': "",
    'global': ""
}

#
#   Synaptic delays
#
delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        int delay""",
        'init': "delay = delays[0][0];",
        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    }
}

#
# Implement the continuous transmission for rate-code synapses.
#
rate_psp_kernel = {
    'body': {
        'sum': ""        
    },
    'header': "",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        // ELLPACK - partition
        int nb_dendrites = proj%(id_proj)s->nb_dendrites();
        int nb_blocks = static_cast<int>(ceil(static_cast<double>(nb_dendrites)/32.0));
        cu_proj%(id_proj)s_psp_ell<<< nb_blocks, 32>>>(
                       nb_dendrites,
                       /* ranks and offsets */
                       proj%(id_proj)s->get_ell()->gpu_post_ranks_, proj%(id_proj)s->get_ell()->gpu_col_idx_, proj%(id_proj)s->get_ell()->get_maxnzr(), proj%(id_proj)s->get_ell()->zero_marker()
                       /* computation data */
                       %(add_args_ell)s
                       /* result */
                       %(target_arg)s );

    #ifdef _DEBUG
        auto ell_err = cudaGetLastError();
        if ( ell_err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp (ELL-partition): " << cudaGetErrorString(ell_err) << std::endl;
        }
    #endif

        // Coordinate - partition
        size_t nb_coo_synapses = proj%(id_proj)s->get_coo()->nb_synapses();
        // check if there is something to compute ...
        if (nb_coo_synapses > 0) {
            int sharedMemSize = proj%(id_proj)s->get_coo()->segment_size() * sizeof(%(float_prec)s);
            cu_proj%(id_proj)s_psp_coo<<< proj%(id_proj)s->get_coo()->number_of_segments(), proj%(id_proj)s->_threads_per_block, sharedMemSize >>>(
                proj%(id_proj)s->get_coo()->segment_size(), proj%(id_proj)s->get_coo()->gpu_segments(), proj%(id_proj)s->get_coo()->gpu_row_indices(), proj%(id_proj)s->get_coo()->gpu_column_indices()
                /* other variables */
                %(add_args_coo)s
                /* result */
                %(target_arg)s
            );

        #ifdef _DEBUG
            auto err_coo = cudaGetLastError();
            if ( err_coo != cudaSuccess ) {
                std::cout << "cu_proj%(id_proj)s_psp (COO-partition): " << cudaGetErrorString(err_coo) << std::endl;
            }
        #endif
        }
    }    
""",
    'kernel_call': "",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }    
}

conn_templates = {
    # connectivity representation
    'conn_header': None,    #constructed from ELL+COO
    'conn_call': None,      #constructed from ELL+COO
    'conn_kernel': None,

    # launch config
    'launch_config': launch_config,
 
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,

    # operations
    'rate_psp': rate_psp_kernel
}


import sympy as sp
from sympy.parsing.sympy_parser import parse_expr, standard_transformations, convert_xor, auto_number
from sympy.printing.latex import LatexPrinter
import re

from ANNarchy.intern import Messages
from ANNarchy.core.Random import RandomDistribution
from ..Extraction import *
from ANNarchy.parser.AnalyseSynapse import analyse_synapse


##################################
### Process individual equations
##################################

def _process_random(val):
    "Transforms a connector attribute (weights, delays) into a string representation"
    if isinstance(val, RandomDistribution):
        return val.latex()
    else:
        return str(val)

# Really crappy...
# When target has a number (ff1), sympy thinks the 1 is a number
# the target is replaced by a text to avoid this
target_replacements = [
    'firsttarget',
    'secondtarget',
    'thirdtarget',
    'fourthtarget',
    'fifthtarget',
    'sixthtarget',
    'seventhtarget',
    'eighthtarget',
    'ninthtarget',
    'tenthtarget',
    'eleventhtarget',
    'twefthtarget',
    'thirteenthtarget',
    'fourteenthtarget'
]

def _process_neuron_equations(neuron, net_id):

    # Extract parameters and variables
    parameters = extract_parameters(neuron.parameters, neuron.extra_values, 'neuron', net_id=net_id)
    variables = extract_variables(neuron.equations, 'neuron', net_id=net_id)
    variable_names = [var['name'] for var in variables]
    attributes, local_var, semiglobal_var, global_var = get_attributes(parameters, variables, neuron=True)

    # Create a dictionary for parsing
    local_dict = {
        'g_target': sp.Symbol(r'g_{\text{target}}'),
        'dt': sp.Symbol(r'\Delta t'),
        't_pre': sp.Symbol(r't_{\text{pre}}'),
        't_post': sp.Symbol(r't_{\text{pos}}'),
        'Uniform': sp.Function(r'\mathcal{U}'),
        'Normal': sp.Function(r'\mathcal{N}'),
        'ite': sp.Function(r'ite', nargs=3)
    }

    for att in attributes:
        local_dict[att] = sp.Symbol(_latexify_name(att, variable_names))

    tex_dict = {}
    for key, val in local_dict.items():
        tex_dict[val] = str(val)

    for var in variables:
        # Retrieve the equation
        eq = var['eq']

        # Extract sum(target)
        targets = []
        target_list = re.findall(r'(?P<pre>[^\w.])sum\(\s*([^()]+)\s*\)', eq)
        for l, t in target_list:
            if t.strip() == '':
                continue
            replacement = target_replacements[len(targets)]
            targets.append((t.strip(), replacement))
            local_dict[replacement] = sp.Symbol(replacement)
            tex_dict[replacement] = replacement

        for target, repl in targets:
            eq = eq.replace('sum('+target+')', repl)

        # Parse the equation
        ode = re.findall(r'([^\w]*)d([\w]+)/dt', eq)
        if len(ode) > 0:
            name = ode[0][1]
            
            eq = eq.replace('d'+name+'/dt', '_grad_'+name)
            
            grad_symbol = sp.Symbol(r'\frac{d'+_latexify_name(name, variable_names)+'}{dt}')
            
            local_dict['_grad_'+name] = grad_symbol
            
            tex_dict[grad_symbol] = r'\frac{d'+_latexify_name(name, variable_names)+'}{dt}'

        var_code = _analyse_equation(var['eq'], eq, local_dict, tex_dict)

        # Replace the targets
        for target, repl in targets:
            target = target.replace("_", r"\_")
            var_code = var_code.replace(
                repl, 
                r'\sum_{\text{'+target+r'}} w \cdot r^{\text{pre}}(t-d)'
            )

        # Add the code
        var['latex'] = var_code
        var['ode'] = len(ode) > 0

    if not neuron.spike: # rate-code, no spike
        return variables, "", []

    # Additional code for spiking neurons
    spike_condition = _analyse_part(neuron.spike, local_dict, tex_dict)

    # Reset
    spike_reset = []
    if neuron.reset is not None:
        reset_vars = extract_variables(neuron.reset, 'neuron', net_id=net_id)
        for var in reset_vars:
            eq = var['eq']
            spike_reset.append(_analyse_equation(var['eq'], eq, local_dict, tex_dict))

    return variables, spike_condition, spike_reset


def _process_synapse_equations(synapse, net_id):
    psp = ""
    code = ""
    pre_event = []
    post_event = []

    # Extract parameters and variables
    parameters = extract_parameters(synapse.parameters, object_type='synapse', net_id=net_id)
    variables = extract_variables(synapse.equations, 'synapse', net_id=net_id)
    variable_names = [var['name'] for var in variables]
    attributes, local_var, semiglobal_var, global_var = get_attributes(parameters, variables, neuron=False)

    # Create a dictionary for parsing
    local_dict = {
        'w': sp.Symbol(r'w(t)'),
        'dt': sp.Symbol(r'\Delta t'),
        'g_target': sp.Symbol(r'g_{\text{target}(t)}'),
        't_pre': sp.Symbol(r't_{\text{pre}}'),
        't_post': sp.Symbol(r't_{\text{pos}}'),
        'Uniform': sp.Function(r'\mathcal{U}'),
        'Normal': sp.Function(r'\mathcal{N}'),
        'ite': sp.Function(r'ite', nargs=3)
    }

    for att in attributes:
        local_dict[att] = sp.Symbol(_latexify_name(att, variable_names))

    tex_dict = {}
    for key, val in local_dict.items():
        tex_dict[val] = str(val)


    # PSP
    if synapse.psp:
        psp, untouched_var, dependencies = extract_prepost('psp', synapse.psp.strip(), synapse.description, net_id=net_id)
        for dep in dependencies['post']:
            local_dict['_post_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{post}}}(t)")
        for dep in dependencies['pre']:
            local_dict['_pre_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{pre}}}(t-d)")
        if synapse.type == 'rate':  
            psp = _analyse_part(psp, local_dict, tex_dict)
        else:  
            psp = r"g_\text{target}(t) \mathrel{+}= " + _analyse_part(psp, local_dict, tex_dict)
    else:
        if synapse.type == 'rate':
            psp = r"w(t) \cdot r^{\text{pre}}(t)"
        else:
            psp = ""


    # Variables
    for var in variables:
        # Retrieve the equation
        eq = var['eq']

        # pre/post variables
        targets=[]
        eq, untouched_var, dependencies = extract_prepost(var['name'], eq, synapse.description, net_id=net_id)

        for dep in dependencies['post']:
            if dep.startswith('sum('):
                target = re.findall(r'sum\(([\w]+)\)', dep)[0]
                targets.append(target)
                local_dict['_post_sum_'+target] = sp.Symbol('PostSum'+target)
            else:
                content = r"{" + _latexify_name(dep, variable_names) + r"^{\text{post}}}(t)"
                local_dict['_post_'+dep+'__'] = sp.Symbol(content)

        for dep in dependencies['pre']:
            if dep.startswith('sum('):
                target = re.findall(r'sum\(([\w]+)\)', dep)[0]
                targets.append(target)
                local_dict['_pre_sum_'+target] = sp.Symbol('PreSum'+target)
            else:
                local_dict['_pre_'+dep+'__'] = sp.Symbol("{" + dep + r"^{\text{pre}}}(t-d)")

        # Parse the equation
        #eq = eq.replace(' ', '') # supress spaces
        ode = re.findall(r'([^\w]*)d([\w]+)/dt', eq)
        if len(ode) > 0:
            
            name = ode[0][1]
            eq = eq.replace('d'+name+'/dt', '_grad_'+name)
            
            grad_symbol = sp.Symbol(r'\frac{d'+_latexify_name(name, variable_names)+r'}{dt}')
            
            local_dict['_grad_'+name] = grad_symbol
            
            tex_dict[grad_symbol] = r'\frac{d'+_latexify_name(name, variable_names)+r'}{dt}'

        # Analyse
        var_code = _analyse_equation(var['eq'], eq, local_dict, tex_dict)

        # replace targets
        for target in targets:
            
            var_code = var_code.replace('PostSum'+target, r"(\sum_{\text{" + target + r"}} \text{psp}(t))^{\text{post}}")
            
            var_code = var_code.replace('PreSum'+target,  r"(\sum_{\text{" + target + r"}} \text{psp}(t))^{\text{pre}}")

        # Add the code
        var['latex'] = var_code
        var['ode'] = len(ode) > 0

    # Pre-event
    if synapse.type == 'spike':
        desc = analyse_synapse(synapse, net_id=net_id)
        for var in extract_pre_spike_variable(desc, net_id=net_id):
            eq = var['eq']
            # pre/post variables
            eq, untouched_var, dependencies = extract_prepost(var['name'], eq, desc, net_id=net_id)
            for dep in dependencies['post']:
                local_dict['_post_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{post}}}(t)")
            for dep in dependencies['pre']:
                local_dict['_pre_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{pre}}}(t)")

            pre_event.append(_analyse_equation(var['eq'], eq, local_dict, tex_dict))

        for var in extract_post_spike_variable(desc, net_id=net_id):
            eq = var['eq']
            # pre/post variables
            eq, untouched_var, dependencies = extract_prepost(var['name'], eq, desc, net_id=net_id)
            for dep in dependencies['post']:
                local_dict['_post_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{post}}}(t)")
            for dep in dependencies['pre']:
                local_dict['_pre_'+dep+'__'] = sp.Symbol(r"{" + dep + r"^{\text{pre}}}(t)")

            post_event.append(_analyse_equation(var['eq'], eq, local_dict, tex_dict))

    return psp, variables, pre_event, post_event


def _process_functions(functions, begin="\\begin{dmath*}\n", end="\n\\end{dmath*}", net_id=0):
    code = ""

    extracted_functions = extract_functions(description=functions, local_global=False, net_id=net_id)

    for func in extracted_functions:
        # arguments
        args = func['args']
        args_list = ""
        for arg in args:
            args_list += _latexify_name(arg, []) + ", "
        args_list = args_list[:-2]

        # local dict
        local_dict = {}
        for att in args:
            local_dict[att] = sp.Symbol(_latexify_name(att, []))
        tex_dict = {}
        for key, val in local_dict.items():
            tex_dict[val] = str(val)

        # parse the content
        content = _analyse_part(func['content'], local_dict, tex_dict)

        # generate the code
        code += "%(begin)s%(name)s(%(args)s) = %(content)s%(end)s" %  {
            'name': _latexify_name(func['name'], []), 
            'args': args_list, 
            'content': content.strip(), 
            'begin': begin, 
            'end': end }


    return code


# Splits an equation into two parts, caring for the increments
def _analyse_equation(orig, eq, local_dict, tex_dict):

    # Analyse the left part
    left = eq.split('=')[0]
    split_idx = len(left)
    if left[-1] in ['+', '-', '*', '/']:
        op = left[-1]
        try:
            left = _analyse_part(left[:-1], local_dict, tex_dict)
        except Exception as e:
            Messages._print(e)
            Messages._warning('can not transform the left side of ' + orig +' to LaTeX, you have to do it by hand...')
            left = left[:-1]
        operator = " = " + left +  " " + op + (" (" if op != '+' else '')
        operator = r" \mathrel{" + op + r"}= " 
    else:
        try:
            left = _analyse_part(left, local_dict, tex_dict)
        except Exception as e:
            Messages._print(e)
            Messages._warning('can not transform the left side of ' + orig +' to LaTeX, you have to do it by hand...')
        operator = " = "

    # Analyse the right part
    try:
        right = _analyse_part(eq[split_idx+1:], local_dict, tex_dict)
    except Exception as e:
        Messages._print(e)
        Messages._warning('can not transform the right side of ' + orig +' to LaTeX, you have to do it by hand...')
        right = r"\textbf{TODO} %%" + eq[split_idx+1:]

    return left + operator + right + (" )" if operator.strip().endswith('(') else "")

class CustomLatexPrinter(LatexPrinter):
    def _print_Function(self, expr, exp=None):
        '''
        For ite(), pos() and neg() only.
        '''
        func = expr.func.__name__
        args = [ str(self._print(arg)) for arg in expr.args ]
        
        if func == 'ite':
            return r"""
\begin{cases}
    %(then_code)s \qquad \text{if} \quad %(if_code)s \\
    \\
    %(else_code)s \qquad \text{otherwise.} 
\end{cases}""" % {'if_code': args[0], 'then_code': args[1], 'else_code': args[2]}

        elif func in ['positive', 'pos']:
            return r"(" + args[0] + r")^+"
        
        elif func in ['negative', 'neg']:
            return r"(" + args[0] + r")^-"

        return LatexPrinter._print_Function(self, expr, exp)

# Analyses and transform to latex a single part of an equation
def _analyse_part(expr, local_dict, tex_dict):

    def regular_expr(expr):
        analysed = parse_expr(
            expr,
            local_dict = local_dict,
            transformations = (standard_transformations + (convert_xor,)),
            # transformations = (convert_xor,),
            # evaluate=False
            )
        return CustomLatexPrinter(settings={'symbol_names': tex_dict, 'mul_symbol':"dot"}).doprint(analysed)

    def _condition(condition):
        return regular_expr(transform_condition(condition))

    def _extract_conditional(condition):
        if_statement = condition[0]
        then_statement = condition[1]
        else_statement = condition[2]
        # IF condition
        if_code = _condition(if_statement)
        # THEN
        if isinstance(then_statement, list): # nested conditional
            then_code =  _extract_conditional(then_statement)
        else:
            then_code = regular_expr(then_statement)
        # ELSE
        if isinstance(else_statement, list): # nested conditional
            else_code =  _extract_conditional(else_statement)
        else:
            else_code = regular_expr(else_statement)
        return r"\begin{cases}" + then_code + r"\qquad \text{if} \quad " + if_code + r"\\ "+ else_code + r" \qquad \text{otherwise.} \end{cases}"

    # Replace and/or with sympy relationals
    expr = transform_condition(expr)

    # Extract if/then/else
    if 'else' in expr:
        ite_code = extract_ite(expr)
        return _extract_conditional(ite_code)

    # Return the transformed equation
    return regular_expr(expr)

def extract_ite(eq):
    "if COND: THEN else: ELSE"

    def transform(code):
        " Transforms the code into a list of lines."
        res = []
        items = []
        for arg in code.split(':'):
            items.append( arg.strip())
        for i in range(len(items)):
            if items[i].startswith('if '):
                res.append( items[i].strip() )
            elif items[i].strip().endswith('else'):
                res.append(items[i].split('else')[0].strip() )
                res.append('else' )
            else: # the last then
                res.append( items[i].strip() )
        return res


    def parse(lines):
        " Recursive analysis of if-else statements"
        result = []
        while lines:
            if lines[0].startswith('if'):
                block = [lines.pop(0).split('if')[1], parse(lines)]
                if lines[0].startswith('else'):
                    lines.pop(0)
                    block.append(parse(lines))
                result.append(block)
            elif not lines[0].startswith(('else')):
                result.append(lines.pop(0))
            else:
                break
        return result[0]

    # Process the equation
    multilined = transform(eq)
    condition = parse(multilined)
    return condition

def extract_ite_func(eq):
    "ite(COND, THEN, ELSE)"
    # Process the equation
    condition = ["r>0.0", "1.0", "0.0"]
    return condition

# Latexify names
greek = ['alpha', 'beta', 'gamma', 'epsilon', 'eta', 'kappa', 'delta', 'lambda', 'mu', 'nu', 'zeta', 'sigma', 'phi', 'psi', 'rho', 'omega', 'xi', 'tau',
         'Gamma', 'Delta', 'Theta', 'Lambda', 'Xi', 'Phi', 'Psi', 'Omega'
]

def _latexify_name(name, local):
    parts = name.split('_')
    if len(parts) == 1:
        if len(name) == 1:
            equiv = name
        elif name in greek:
            equiv = '\\' + name
        else:
            equiv = r'{\text{' + name + '}}'
        if name in local:
            equiv = '{' + equiv + '}(t)'
        return equiv
    elif len(parts) == 2:
        equiv = ""
        for p in parts:
            if len(p) == 1:
                equiv += '' + p + '_'
            elif p in greek:
                equiv += '\\' + p + '_'
            else:
                equiv += r'{\text{' + p + '}}' + '_'
        equiv = equiv[:-1]
        if name in local:
            equiv = '{' + equiv + '}(t)'
        return equiv
    else:
        equiv = r'{\text{' + name + '}}'
        equiv = equiv.replace('_', '-')
        if name in local:
            equiv = equiv + '(t)'
        return equiv

def pop_name(name):
    return name.replace('_', r'\_')


def _format_list(l, sep):
    if not isinstance(l, list):
        return l
    target_list = ""
    for t in l:
        target_list += t + sep
    return target_list[:-len(sep)]

def transform_condition(expr):
    """
    Transforms the "natural" logical operators into Sympy-compatible versions.
    """
    expr = expr.replace (' and ', ' & ')
    expr = expr.replace (' or ', ' | ')
    expr = expr.replace (' is not ', ' != ')
    expr = expr.replace (' not ', ' Not ')
    expr = expr.replace (' not(', ' Not(')
    expr = expr.replace (' is ', ' == ')

    return expr



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import re
from math import ceil
from copy import deepcopy

import ANNarchy

from ANNarchy.generator.Template.GlobalOperationTemplate import global_operation_templates_cuda as global_op_template
from ANNarchy.generator.Population import CUDATemplates
from ANNarchy.generator.Utils import generate_equation_code, tabify, check_and_apply_pow_fix
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from .PopulationGenerator import PopulationGenerator
from .CUDATemplates import cuda_templates

class CUDAGenerator(PopulationGenerator):
    """
    Generate the header for a Population object to use on CUDA devices.
    """
    def __init__(self, cuda_version, profile_generator, net_id):
        super(CUDAGenerator, self).__init__(profile_generator, net_id)
        self._cuda_version = cuda_version

    def header_struct(self, pop, annarchy_dir):
        """
        Specialized implementation of PopulationGenerator.header_struct() for
        generation of a CUDA header.
        """
        self._templates = deepcopy(cuda_templates)

        # Generate declaration and accessors of all parameters and variables
        declaration_parameters_variables, access_parameters_variables = self._generate_decl_and_acc(pop)

        # Additional includes and structures
        include_additional = ""
        struct_additional = ""
        declare_additional = ""
        init_additional = ""
        reset_additional = ""
        access_additional = ""

        # Initialize parameters and variables
        init_parameters_variables = self._init_population(pop)

        # Reset of device to host flags
        reset_read_flags = self._reset_read_flags(pop)

        # Spike-specific stuff
        reset_spike = ""
        declare_spike = ""
        init_spike = ""
        if pop.neuron_type.description['type'] == 'spike':
            spike_tpl = self._templates['spike_specific']

            # Main data for spiking pops
            declare_spike += spike_tpl['spike']['declare'] % {'id': pop.id}
            init_spike += spike_tpl['spike']['init'] % {'id': pop.id}
            reset_spike += spike_tpl['spike']['reset'] % {'id': pop.id}

            # If there is a refractory period
            if pop.neuron_type.refractory or pop.refractory:
                declare_spike += spike_tpl['refractory']['declare'] % {'id': pop.id}
                if isinstance(pop.neuron_type.description['refractory'], str): # no need to instantiate refractory
                    init_spike += spike_tpl['refractory']['init_extern'] % {'id': pop.id}
                else:
                    init_spike += spike_tpl['refractory']['init'] % {'id': pop.id}
                reset_spike += spike_tpl['refractory']['reset'] % {'id': pop.id}

        # Process eventual delay
        declare_delay = ""; init_delay = ""; update_delay = ""; update_max_delay = ""; reset_delay = ""
        if pop.max_delay > 1:
            declare_delay, init_delay, update_delay, update_max_delay, reset_delay = self._delay_code(pop)

        # Process mean FR computations
        declare_FR, init_FR, reset_FR = self._init_fr(pop)
        reset_spike += reset_FR

        update_FR = self._update_fr(pop)

        # Update random distributions
        update_rng = self._update_random_distributions(pop)

        # Update global operations
        update_global_ops = self._update_globalops(pop)

        # Update the neural variables
        if pop.neuron_type.type == "rate":
            body, invoke, header, call = self._update_rate_neuron(pop)
        else:
            update_body, update_invoke, update_header, update_call = self._update_spiking_neuron(pop)
            spike_body, spike_invoke, spike_header, spike_call = self._spike_gather(pop)
            body = update_body + spike_body
            invoke = update_invoke + spike_invoke
            header = update_header + spike_header
            call = update_call + spike_call
        update_variables = ""

        # Memory transfers
        host_to_device, device_to_host = self._memory_transfers(pop)

        # Stop condition
        stop_condition = self._stop_condition(pop)

        # Local functions
        host_local_func, device_local_func = self._local_functions(pop)
        declaration_parameters_variables += host_local_func

        # Memory management
        size_in_bytes = self._size_in_bytes(pop)
        clear_container = self._clear_container(pop)

        # Profiling
        if self._prof_gen:
            include_profile = """#include "Profiling.hpp"\n"""
            declare_profile, init_profile = self._prof_gen.generate_init_population(pop)
        else:
            include_profile = ""
            init_profile = ""
            declare_profile = ""

        ## When everything is generated, we override the fields defined by the specific population
        if 'include_additional' in pop._specific_template.keys():
            include_additional = pop._specific_template['include_additional']
        if 'struct_additional' in pop._specific_template.keys():
            struct_additional = pop._specific_template['struct_additional']
        if 'declare_spike_arrays' in pop._specific_template.keys():
            declare_spike = pop._specific_template['declare_spike_arrays']
        if 'declare_parameters_variables' in pop._specific_template.keys():
            declaration_parameters_variables = pop._specific_template['declare_parameters_variables']
        if 'declare_additional' in pop._specific_template.keys():
            declare_additional = pop._specific_template['declare_additional']
        if 'declare_FR' in pop._specific_template.keys():
            declare_FR = pop._specific_template['declare_FR']
        if 'declare_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            declare_delay = pop._specific_template['declare_delay']
        if 'access_parameters_variables' in pop._specific_template.keys():
            access_parameters_variables = pop._specific_template['access_parameters_variables']
        if 'access_additional' in pop._specific_template.keys():
            access_additional = pop._specific_template['access_additional']
        if 'init_parameters_variables' in pop._specific_template.keys():
            init_parameters_variables = pop._specific_template['init_parameters_variables']
        if 'init_spike' in pop._specific_template.keys():
            init_spike = pop._specific_template['init_spike']
        if 'init_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            init_delay = pop._specific_template['init_delay']
        if 'init_FR' in pop._specific_template.keys():
            init_FR = pop._specific_template['init_FR']
        if 'init_additional' in pop._specific_template.keys():
            init_additional = pop._specific_template['init_additional']
        if 'reset_spike' in pop._specific_template.keys():
            reset_spike = pop._specific_template['reset_spike']
        if 'reset_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            reset_delay = pop._specific_template['reset_delay']
        if 'reset_read_flags' in pop._specific_template.keys():
            reset_read_flags = pop._specific_template['reset_read_flags']
        if 'reset_additional' in pop._specific_template.keys():
            reset_additional = pop._specific_template['reset_additional']
        if 'update_variables' in pop._specific_template.keys():
            update_variables = pop._specific_template['update_variables']
        if 'update_rng' in pop._specific_template.keys():
            update_rng = pop._specific_template['update_rng']
        if 'update_FR' in pop._specific_template.keys():
            update_FR = pop._specific_template['update_FR']
        if 'update_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            update_delay = pop._specific_template['update_delay']
        if 'update_max_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            update_max_delay = pop._specific_template['update_max_delay']
        if 'update_global_ops' in pop._specific_template.keys():
            update_global_ops = pop._specific_template['update_global_ops']

        # Fill the template
        code = self._templates['population_header'] % {
            # version tag
            'annarchy_version': ANNarchy.__release__,
            # fill code templates
            'float_prec': ConfigManager().get('precision', self._net_id),
            'id': pop.id,
            'name': pop.name,
            'size': pop.size,
            'include_additional': include_additional,
            'include_profile': include_profile,
            'struct_additional': struct_additional,
            'extern_global_operations': "", # CPU side global ops
            'declare_spike_arrays': declare_spike,
            'declare_parameters_variables': declaration_parameters_variables,
            'declare_additional': declare_additional,
            'declare_delay': declare_delay,
            'declare_FR': declare_FR,
            'declare_profile': declare_profile,
            'access_parameters_variables': access_parameters_variables,
            'access_additional': access_additional,
            'init_parameters_variables': init_parameters_variables,
            'init_spike': init_spike,
            'init_delay': init_delay,
            'init_FR': init_FR,
            'init_additional': init_additional,
            'init_profile': init_profile,
            'reset_spike': reset_spike,
            'reset_delay': reset_delay,
            'reset_additional': reset_additional,
            'reset_read_flags': reset_read_flags,
            'update_FR': update_FR,
            'update_variables': update_variables,
            'update_rng': update_rng,
            'update_delay': update_delay,
            'update_max_delay': update_max_delay,
            'update_global_ops': update_global_ops,
            'stop_condition': stop_condition,
            'host_to_device': host_to_device,
            'device_to_host': device_to_host,
            'size_in_bytes': size_in_bytes,
            'clear_container': clear_container
        }

        # Store the complete header definition in a single file
        with open(annarchy_dir+'/generate/net'+str(self._net_id)+'/pop'+str(pop.id)+'.hpp', 'w') as ofile:
            ofile.write(code)

        # Basic informations common to all populations
        pop_desc = {
            'include': """#include "pop%(id)s.hpp"\n""" % {'id': pop.id},
            'extern': """extern PopStruct%(id)s* pop%(id)s;\n"""% {'id': pop.id},
            'instance': """PopStruct%(id)s* pop%(id)s;\n"""% {'id': pop.id},
            'init': """    pop%(id)s->init_population();\n""" % {'id': pop.id}
        }

        pop_desc['custom_func'] = device_local_func
        pop_desc['update'] = call
        pop_desc['update_invoke'] = invoke
        pop_desc['update_body'] = body
        pop_desc['update_header'] = header
        pop_desc['update_delay'] = """    pop%(id)s->update_delay();\n""" % {'id': pop.id} if pop.max_delay > 1 else ""
        pop_desc['update_FR'] = """    pop%(id)s->update_FR();\n""" % {'id': pop.id} if pop.neuron_type.type == "spike" else ""

        if len(pop.global_operations) > 0:
            pop_desc['gops_update'] = self._update_globalops(pop) % {'id': pop.id}

        pop_desc['host_to_device'] = tabify("pop%(id)s->host_to_device();" % {'id':pop.id}, 1)+"\n"

        return pop_desc

    def _clear_container(self, pop):
        """
        Clear allocated data structures.

        The function overrrides the default behavior as we need a de-allocation
        on host and device side.
        """
        from ANNarchy.generator.Utils import tabify

        # Variables (host-side), which contains also Mean-FR code
        code = PopulationGenerator._clear_container(self, pop)

        # Variables (device side)
        code += "// parameters\n"
        for attr in pop.neuron_type.description['parameters']:
            if attr['locality'] == "local":
                code += """cudaFree(gpu_%(name)s); \n""" % {'name': attr['name']}

        code += "\n// variables\n"
        for attr in pop.neuron_type.description['variables']:
            code += """cudaFree(gpu_%(name)s); \n""" % {'name': attr['name']}

        code += "\n// delayed attributes\n"
        if pop.neuron_type.type == "rate":
            delay_tpl = self._templates['attribute_delayed']
            for var in pop.delayed_variables:
                if var in pop.neuron_type.description['local']:
                    code += delay_tpl['local']['clear'] % {'name': var}
                else:
                    continue

        # clear PSP targets ( for rate-coded neurons, for spiking they are part of population variables )
        code += "\n// targets\n"
        if pop.neuron_type.type == 'rate':
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                code += """cudaFree(gpu__sum_%(target)s); \n""" % {'target': target}

        # just code layout
        code = tabify(code, 2)

        # Sanitiy check
        code += """
    #ifdef _DEBUG
        cudaError_t err_clear = cudaGetLastError();
        if ( err_clear != cudaSuccess )
            std::cout << "Pop%(id)s::clear() - cudaFree: " << cudaGetErrorString(err_clear) << std::endl;
    #endif
""" % {'id': pop.id}

        # code complete
        return code

    def reset_computesum(self, pop):
        """
        For rate-coded networks we need to reset the weighted sum in each step, as the kernels simply
        add up.
        """
        if pop.neuron_type.type != 'rate':
            return ""

        code = ""
        for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
            code += """
    #if defined (__pop%(id)s_nb__)
        call_clear_sum( RunConfig(__pop%(id)s_nb__, __pop%(id)s_tpb__, 0, pop%(id)s->stream), pop%(id)s->size, pop%(id)s->gpu__sum_%(target)s );
    #else
        call_clear_sum( RunConfig(pop%(id)s->_nb_blocks, pop%(id)s->_threads_per_block, 0, pop%(id)s->stream), pop%(id)s->size, pop%(id)s->gpu__sum_%(target)s );
    #endif
""" % {'id': pop.id, 'target': target}

        if code != "":
            code = """
    if ( pop%(id)s->_active ) {
%(reset_code)s
    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "clear_sum: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""" % {'reset_code': code, 'id': pop.id}

        return code

    def _reset_read_flags(self, pop):
        """
        Reset of device to host flags.

        The read-back from the GPUs happens if the first time is accessed the variable. To know, if the GPU simulated
        between two subsequent calls of such an getter, we store the time stamp of the last read-back. This must be set
        back to 0 in case of a reset.
        """
        reset_read_flags = "// read-back flags: variables\n"

        attributes = []
        for var in pop.neuron_type.description["variables"]:
            if var['name'] in attributes:
                continue

            reset_read_flags += var['name']+"_device_to_host = 0;\n"
            attributes.append(var['name'])

        reset_read_flags += "\n// read-back flags: targets\n"
        if pop.neuron_type.type == 'rate':
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                reset_read_flags += "_sum_"+target+"_device_to_host = 0;\n"

        return tabify(reset_read_flags, 2)

    def _delay_code(self, pop):
        """
        Generate code for delayed variables, comprising of initialization
        and update codes.

        Implementation Note: (HD: 15.06.2015)

            Currently I see no better way to implement delays, as consequence of missing device-device memory transfers ...

            This implementation is from a performance point of view problematic, cause of low host-device memory bandwith,
            maybe enhancable through pinned memory (CC 2.x), or asynchronous device transfers (CC 3.x)

        Algorithm:

            * get reference of last list in queue (last_%(var)s)
            * cycle the last_%(var)s pointer from back to front
            * get current value from %(var)s and store it in tmp_%(var)s
            * store the data in queue
        """
        delay_tpl = self._templates['attribute_delayed']

        # Declaration
        declare_code = """
    // Delayed variables"""
        if pop.neuron_type.type == "rate":
            for var in pop.delayed_variables:
                attr = self._get_attr(pop, var)
                declare_code += delay_tpl[attr['locality']]['declare'] % {'var': var, 'type': attr['ctype']}
        else:
            # Spiking networks should only exchange spikes
            declare_code += """
    // Delays for spike population
    std::deque< int* > gpu_delayed_spiked;              // contains a set of device pointers
    std::deque< unsigned int > host_delayed_num_events; // number of events stored in each container
"""
            # TODO:
            if pop.delayed_variables != []:
                raise NotImplementedError

        # Initialization
        init_code = """
        // Delayed variables"""

        # Update and Reset for delayed variables
        update_code = ""
        reset_code = ""
        resize_code = ""
        for var in pop.delayed_variables:
            attr = self._get_attr(pop, var)

            attr_dict = {
                'id': pop.id,
                'name': attr['name'],
                'type': attr['ctype'],
                'delay': pop.max_delay
            }

            init_code += delay_tpl[attr['locality']]['init'] % attr_dict
            update_code += delay_tpl[attr['locality']]['update'] % attr_dict
            reset_code += delay_tpl[attr['locality']]['reset'] % attr_dict
            resize_code += delay_tpl[attr['locality']]['resize'] % attr_dict

        # Delaying spike events is done differently
        if pop.neuron_type.type == 'spike':
            init_code += """
            gpu_delayed_spiked = std::deque<int*>();
            host_delayed_num_events = std::deque<unsigned int>();
            int *dev_spiked;

            for(int i = 0; i < max_delay; i++) {
                // we allocate max size
                cudaMalloc((void**)&dev_spiked, size * sizeof(int));
                gpu_delayed_spiked.push_front(dev_spiked);

                // true number of events stored in dev_spiked
                host_delayed_num_events.push_front(static_cast<unsigned int>(0));
            }
            """ % {'max_delay': int(ceil(pop.max_delay/ConfigManager().get('dt', self._net_id)))}
            update_code += """
            // Take the last vector in queue and push it to front
            int* last_spiked = gpu_delayed_spiked.back();
            gpu_delayed_spiked.pop_back();
            gpu_delayed_spiked.push_front(last_spiked);

            // do not copy empty vectors!
            if (spike_count > 0) {
                cudaMemcpy( last_spiked, gpu_spiked, spike_count * sizeof(int), cudaMemcpyDeviceToDevice);

            #ifdef _DEBUG
                auto err1 = cudaGetLastError();
                if (err1 != cudaSuccess) {
                    std::cerr << "PopStruct%(id)s::update_delay() - spiked :" << cudaGetErrorString(err1) << std::endl;
                }
            #endif
            }

            host_delayed_num_events.pop_back();
            host_delayed_num_events.push_front(spike_count);

        #ifdef _DEBUG
            std::cout << "PopStruct::update_delay() at t = " << t << std::endl;
            std::cout << "[";
            for (auto i = 0; i < host_delayed_num_events.size(); i++)
                std::cout << host_delayed_num_events[i] << ", ";
            std::cout << "]" << std::endl;
        #endif
            """ % {'id': pop.id}
            reset_code += ""

        update_code = """
        if ( _active ) {
%(code)s
        }""" % {'code': update_code}

        return declare_code, init_code, update_code, resize_code, reset_code

    def _init_fr(self, pop):
        """
        Declares arrays for computing the mean FR of a spiking neuron.

        HD ( 09. March 2017 ):

            As a queue is hard to realize on the device,
            we do the computation on the CPU - side for now.
        """
        declare_FR = ""; init_FR = ""; reset_FR = ""
        if pop.neuron_type.description['type'] == 'spike':
            declare_FR = """
    // Mean Firing rate
    std::vector< std::queue<long int> > _spike_history;
    long int _mean_fr_window;
    %(float_prec)s _mean_fr_rate;
    void compute_firing_rate( %(float_prec)s window){
        if(window>0.0){
            _mean_fr_window = int(window/dt);
            _mean_fr_rate = %(float_prec)s(1000./%(float_prec)s(window));
            if (_spike_history.empty())
                _spike_history = std::vector< std::queue<long int> >(size, std::queue<long int>());
        }
    };""" % {'float_prec': ConfigManager().get('precision', self._net_id)}
            init_FR = """
        // Mean Firing Rate
        _spike_history = std::vector< std::queue<long int> >();
        _mean_fr_window = 0;
        _mean_fr_rate = 1.0;"""
            reset_FR = """
        // Mean Firing Rate
        for (auto it = _spike_history.begin(); it != _spike_history.end(); it++) {
            if (!it->empty()) {
                auto empty_queue = std::queue<long int>();
                it->swap(empty_queue);
            }
        }
"""

        return declare_FR, init_FR, reset_FR

    def _gen_kernel_args(self, pop, locality):
        """
        Generate the argument and call statemen for neural variables
        used in equations as well as there dependencies.
        """
        # Gather all variable names
        add_args_header = "const long int t, const %(type)s dt" % {'type':ConfigManager().get('precision', self._net_id)}
        add_args_invoke = "t, dt"
        add_args_call = "t, dt"

        deps = []

        # Variables
        for var in pop.neuron_type.description['variables']:
            if var['locality'] == locality:
                deps.append(var['name'])
                if 'dependencies' in var.keys():
                    deps += var['dependencies']

        # Random distributions
        for rd in pop.neuron_type.description['random_distributions']:
            for dep in rd['dependencies']:
                deps.append(dep)

        # Remove doublons / sort the attribute names to
        # prevent unnecessary recompiles
        deps = sorted(list(set(deps)))

        # Generate the header and call lines
        for dep in deps:
            attr_type, attr_dict = self._get_attr_and_type(pop, dep)
            if attr_type is None:
                continue

            ids = {
                'id': pop.id,
                'name': attr_dict['name'],
                'type': attr_dict['ctype']
            }

            if attr_type == 'par':
                if dep in pop.neuron_type.description['global']:
                    add_args_header += ", const %(type)s %(name)s" % ids
                    add_args_invoke += ", %(name)s" % ids
                    add_args_call += ", pop%(id)s->%(name)s" % ids
                else:
                    add_args_header += ", %(type)s* __restrict__ %(name)s" % ids
                    add_args_invoke += ", %(name)s" % ids
                    add_args_call += ", pop%(id)s->gpu_%(name)s" % ids
            elif attr_type == 'var':
                add_args_header += ", %(type)s* __restrict__ %(name)s" % ids
                add_args_invoke += ", %(name)s" % ids
                add_args_call += ", pop%(id)s->gpu_%(name)s" % ids
            elif attr_type == 'rand':
                add_args_header += ", curandState* state_%(name)s" % ids
                add_args_invoke += ", state_%(name)s" % ids
                add_args_call += ", pop%(id)s->gpu_%(name)s" % ids
            else:
                raise NotImplementedError

        return add_args_header, add_args_invoke, add_args_call

    def _local_functions(self, pop):
        """
        Definition of user-defined local functions attached to
        a neuron. These functions will take place in the
        ANNarchyDevice.cu file.

        As the local functions can be occur repeatadly in the same file,
        there are modified with pop[id]_ to unique them.

        Return:

            * host_define, device_define
        """
        # Local functions
        if len(pop.neuron_type.description['functions']) == 0:
            return "", ""

        host_code = ""
        device_code = ""
        for func in pop.neuron_type.description['functions']:
            cpp_func = func['cpp'] + '\n'

            host_code += cpp_func
            # TODO: improve code
            if (ConfigManager().get('precision', self._net_id) == "float"):
                device_code += cpp_func.replace('float ' + func['name'], '__device__ float pop%(id)s_%(func)s' % {'id': pop.id, 'func': func['name']})
            else:
                device_code += cpp_func.replace('double ' + func['name'], '__device__ double pop%(id)s_%(func)s' % {'id': pop.id, 'func': func['name']})

        return host_code, check_and_apply_pow_fix(device_code, self._cuda_version)

    def _replace_local_funcs(self, pop, glob_eqs, loc_eqs):
        """
        As the local functions can be occur repeatadly in the same file,
        there are modified with pop[id]_ to unique them. Now we need
        to adjust the call accordingly.

        Please note: the placeholder like %(global_idx)s etc. must be already filled.
        """
        for func in pop.neuron_type.description['functions']:
            search_term = r"%(name)s\([^\(]*\)" % {'name': func['name']}

            func_occur = re.findall(search_term, glob_eqs)
            for term in func_occur:
                glob_eqs = loc_eqs.replace(term, term.replace(func['name'], 'pop'+str(pop.id)+'_'+func['name']))

            func_occur = re.findall(search_term, loc_eqs)
            for term in func_occur:
                loc_eqs = loc_eqs.replace(term, term.replace(func['name'], 'pop'+str(pop.id)+'_'+func['name']))

        return glob_eqs, loc_eqs

    def _replace_random(self, loc_eqs, glob_eqs, random_distributions):
        """
        We replace the rand_%(id)s by the corresponding curand... term.

        Update (ANNarchy 4.6.10.1):

            Further we need to ensure to draw the variables before solving the ODEs, otherwise we obtain wrong results for
            higher order solving methods.
        """
        # double precision methods have a postfix
        prec_extension = "" if ConfigManager().get('precision', self._net_id) == "float" else "_double"

        loc_pre = ""
        glob_pre = ""
        pre_loop = ""
        post_loop = ""
        for rd in random_distributions:
            if rd['locality'] == "local":
                term = ""
                if rd['dist'] == "Uniform":
                    term = """( curand_uniform%(postfix)s( &loc_state_%(rd)s ) * (%(max)s - %(min)s) + %(min)s )""" % {'postfix': prec_extension, 'rd': rd['name'], 'min': rd['args'].split(',')[0], 'max': rd['args'].split(',')[1]}
                elif rd['dist'] == "Normal":
                    term = """( curand_normal%(postfix)s( &loc_state_%(rd)s ) * %(sigma)s + %(mean)s )""" % {'postfix': prec_extension, 'rd': rd['name'], 'mean': rd['args'].split(",")[0], 'sigma': rd['args'].split(",")[1]}
                elif rd['dist'] == "LogNormal":
                    term = """( curand_log_normal%(postfix)s( &loc_state_%(rd)s, %(mean)s, %(std_dev)s) )""" % {'postfix': prec_extension, 'rd': rd['name'], 'mean': rd['args'].split(',')[0], 'std_dev': rd['args'].split(',')[1]}
                else:
                    Messages._error("Unsupported random distribution on GPUs: " + rd['dist'])

                # suppress local index
                loc_eqs = loc_eqs.replace(rd['name']+"%(local_index)s", rd['name'])

                # add the init
                loc_pre += "%(prec)s %(name)s = %(term)s;" % {'prec': ConfigManager().get('precision', self._net_id), 'name': rd['name'], 'term': term}

                # read-out/write-back of the RNG state
                pre_loop += f"curandState loc_state_{rd['name']} = state_{rd['name']}[tid];"
                post_loop += f"state_{rd['name']}[tid] = loc_state_{rd['name']};"

            else:
                # For global variables we directly access the RNG state.
                term = ""
                if rd['dist'] == "Uniform":
                    term = """( curand_uniform%(postfix)s( &state_%(rd)s[0] ) * (%(max)s - %(min)s) + %(min)s )""" % {'postfix': prec_extension, 'rd': rd['name'], 'min': rd['args'].split(',')[0], 'max': rd['args'].split(',')[1]}
                elif rd['dist'] == "Normal":
                    term = """( curand_normal%(postfix)s( &state_%(rd)s[0] ) * %(sigma)s + %(mean)s )""" % {'postfix': prec_extension, 'rd': rd['name'], 'mean': rd['args'].split(",")[0], 'sigma': rd['args'].split(",")[1]}
                elif rd['dist'] == "LogNormal":
                    term = """( curand_log_normal%(postfix)s( &state_%(rd)s[0], %(mean)s, %(std_dev)s) )""" % {'postfix': prec_extension, 'rd': rd['name'], 'mean': rd['args'].split(',')[0], 'std_dev': rd['args'].split(',')[1]}
                else:
                    Messages._error("Unsupported random distribution on GPUs: " + rd['dist'])

                # suppress global index
                glob_eqs = glob_eqs.replace(rd['name']+"%(global_index)s", rd['name'])

                # add the init
                glob_pre += "%(prec)s %(name)s = %(term)s;" % {'prec': ConfigManager().get('precision', self._net_id), 'name': rd['name'], 'term': term}

        # check which equation blocks we need to extend
        if len(loc_pre) > 0:
            loc_eqs = tabify(loc_pre, 2) + "\n" + loc_eqs
        if len(glob_pre) > 0:
            glob_eqs = tabify(glob_pre, 1) + "\n" + glob_eqs

        return loc_eqs, glob_eqs, pre_loop, post_loop

    def _stop_condition(self, pop):
        """
        Simulation can either end after a fixed point in time or
        dependent on a population related condition. The code for
        this is generated here and added to the ANNarchy.cpp/.cu
        file.
        """
        if not pop.stop_condition: # no stop condition has been defined
            return ""

        # Special case for early return based on emitted events
        if pop.stop_condition.replace(" ", "") == "spiked:any":
            # HD (3rd May 2023): do not use spiked.empty as for single-thread/openMP
            #                    as this container is pre-allocated to speedup monitor readout.
            stop_code = """
    // Stop condition (any)
    bool stop_condition() {
        return (spike_count>0);
    } """
            return stop_code

        elif pop.stop_condition.replace(" ", "") == "spiked:all":
            stop_code = """
    // Stop condition (all)
    bool stop_condition() {
        return spike_count == size;
    } """
            return stop_code

        # Process the stop condition
        pop.neuron_type.description['stop_condition'] = {'eq': pop.stop_condition}
        from ANNarchy.parser.Extraction import extract_stop_condition
        extract_stop_condition(pop.neuron_type.description, pop.net_id)

        mem_transfer = ""
        for dep in pop.neuron_type.description['stop_condition']['dependencies']:
            attr = self._get_attr(pop, dep)

            if attr['locality'] == "local":
                mem_transfer += """
    cudaMemcpy( %(attr_name)s.data(),  gpu_%(attr_name)s, size * sizeof(%(type)s), cudaMemcpyDeviceToHost);
""" % {'attr_name': attr['name'], 'type': attr['ctype']}

        # Retrieve the code
        condition = pop.neuron_type.description['stop_condition']['cpp']% {
            'id': pop.id,
            'local_index': "[i]",
            'semiglobal_index': '',
            'global_index': ''}

        # Generate the function
        if pop.neuron_type.description['stop_condition']['type'] == 'any':
            stop_code = """
    // Stop condition (any)
    bool stop_condition(){
        %(mem_transfer)s
        for(int i=0; i<size; i++)
        {
            if(%(condition)s){
                return true;
            }
        }
        return false;
    }
    """ % {'condition': condition, 'mem_transfer': mem_transfer}
        else:
            stop_code = """
    // Stop condition (all)
    bool stop_condition(){
        %(mem_transfer)s
        for(int i=0; i<size; i++)
        {
            if(!(%(condition)s)){
                return false;
            }
        }
        return true;
    }
    """ % {'condition': condition, 'mem_transfer': mem_transfer}

        return stop_code

    def _update_fr(self, pop):
        """
        Computes the average firing rate based on history.

        HD ( 09. March 2017 ):

            As a queue is hard to realize on the device,
            we do the computation on the CPU - side for now.
        """
        mean_FR_update = ""
        if pop.neuron_type.description['type'] == 'spike':
            mean_FR_update = """
        if ( _mean_fr_window > 0) {
            // Update the queues
            r_host_to_device = false;

            for ( int i = 0; i < spike_count; i++ ) {
                _spike_history[spiked[i]].push(t);
                r_host_to_device = true; // the queue changed the length
            }

            // Recalculate the mean firing rate
            for (int i = 0; i < size; i++ ) {
                while((_spike_history[i].size() != 0)&&(_spike_history[i].front() <= t - _mean_fr_window)){
                    _spike_history[i].pop(); // Suppress spikes outside the window
                    r_host_to_device = true; // the queue changed the length
                }
                r[i] = _mean_fr_rate * float(_spike_history[i].size());
            }

            // transfer to device
            if ( r_host_to_device ) {
                cudaMemcpy(gpu_r, r.data(), size * sizeof(%(float_prec)s), cudaMemcpyHostToDevice);
                r_host_to_device = false;
            }
        }
"""
        return mean_FR_update % {'float_prec': ConfigManager().get('precision', self._net_id)}

    def _update_globalops(self, pop):
        """
        Update of global functions is a call of pre-implemented
        functions defined in GlobalOperationTemplate. In case of
        CUDA the call semantic will be placed in ANNarchy.cu
        file as part of the host section.
        """
        if len(pop.global_operations) == 0:
            return ""

        code = ""
        for op in pop.global_operations:
            code += global_op_template[op['function']]['call'] % {'id': pop.id, 'type': ConfigManager().get('precision', self._net_id), 'op': op['function'], 'var': op['variable']}

        return code

    def _update_random_distributions(self, pop):
        # HD (27.04.2016):
        # we dont need an update code here, as the drawing of random numbers is done in the Population::step()
        return ""

    def _update_rate_neuron(self, pop):
        """
        Generate the execution code for updating neural variables, more precise local and global ones.

        The resulting code comprise of several parts: creating of local and global update code, generating
        function prototype and finally calling statement.

        Returns:

            * tuple of four code snippets (device_kernel, device_invoke, kernel_decl, host_call)

        """
        # Use pre-defined code template
        if 'update_variables' in pop._specific_template.keys():
            try:
                return pop._specific_template['update_variable_body'], pop._specific_template['update_variable_invoke'], pop._specific_template['update_variable_header'], pop._specific_template['update_variable_call']
            except KeyError:
                Messages._error("\nCode generation error: if one attempts to override the population update on CUDA devices, one need to define all of the following fields of _specific_template dictionary:\n\tupdate_variables, update_variable_call, update_variable_header, update_variable_invoke, update_variable_body")

        # Is there any variable?
        if len(pop.neuron_type.description['variables']) == 0:
            return "", "", "", ""

        device_kernel = ""
        kernel_invoke = ""
        kernel_decl = ""
        local_call = ""
        global_call = ""

        # some defaults
        ids = {
            'id': pop.id,
            'local_index': "[i]",
            'global_index': '[0]'
        }

        # parse the equations
        glob_eqs = generate_equation_code(pop.neuron_type.description, locality='global', padding=1)
        loc_eqs = generate_equation_code(pop.neuron_type.description, locality='local', padding=2)

        # Remove %(global_index)s for global parameters
        for par in pop.neuron_type.description['parameters']:
            if par['locality'] == "global":
                glob_eqs = glob_eqs.replace(par['name']+"%(global_index)s", par['name'])
                loc_eqs = loc_eqs.replace(par['name']+"%(global_index)s", par['name'])

        # Gather pre-loop declaration (dt/tau for ODEs)
        pre_loop = ""
        for var in pop.neuron_type.description['variables']:
            if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                pre_loop += ConfigManager().get('precision', self._net_id) + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'
        if pre_loop.strip() != '':
            pre_loop = """
// Updating the step sizes
""" + pre_loop

        # sum() must generate _sum___all__[i] = _sum_exc[i] + sum_inh[i] + ... at the beginning of local equations
        if '__all__' in pop.neuron_type.description['targets']:
            eqs = " "*8 + "// Sum over all targets\n"
            eqs += " "*8 + "_sum___all__[i] = "
            for target in pop.targets:
                eqs += "_sum_" + target + '[i] + '
            eqs = eqs[:-2]
            eqs += ';\n\n'
            loc_eqs = eqs + loc_eqs

        # replace pow() for SDK < 6.5
        loc_eqs = check_and_apply_pow_fix(loc_eqs, self._cuda_version)
        glob_eqs = check_and_apply_pow_fix(glob_eqs, self._cuda_version)

        # replace the random distributions
        loc_eqs, glob_eqs, rng_pre_loop, rng_post_loop = self._replace_random(loc_eqs, glob_eqs, pop.neuron_type.description['random_distributions'])
        pre_loop += rng_pre_loop

        # Replace %(global_idx)s for global parameters
        for var in pop.neuron_type.description["global"]:
            attr_type, attr_dict = self._get_attr_and_type(pop, var)
            if attr_type == 'par':
                if pre_loop.strip() != '':
                    pre_loop = pre_loop.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if glob_eqs.strip() != '':
                    glob_eqs = glob_eqs.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if loc_eqs.strip() != '':
                    loc_eqs = loc_eqs.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])

        # Fill the placeholder
        glob_eqs = glob_eqs % ids
        loc_eqs = loc_eqs % ids
        pre_loop = pre_loop % ids
        rng_post_loop = rng_post_loop % ids

        # replace local function calls
        if len(pop.neuron_type.description['functions']) > 0:
            glob_eqs, loc_eqs = self._replace_local_funcs(pop, glob_eqs, loc_eqs)

        # Global operations
        if glob_eqs.strip() != '':
            add_args_header, add_args_invoke, add_args_call = self._gen_kernel_args(pop, 'global')

            for op in pop.global_operations:
                ids = {
                    'id': pop.id,
                    'type': ConfigManager().get('precision', self._net_id),
                    'op': op['function'],
                    'var': op['variable']
                }
                add_args_header += """, %(type)s _%(op)s_%(var)s """ % ids
                add_args_call += """, pop%(id)s->_%(op)s_%(var)s""" % ids

            # finalize code templates
            device_kernel += CUDATemplates.population_update_kernel['global']['device_kernel'] % {
                'id': pop.id,
                'add_args': add_args_header,
                'global_eqs':glob_eqs,
                'pre_loop': tabify(pre_loop, 1)
            }
            kernel_invoke += CUDATemplates.population_update_kernel['global']['invoke_kernel'] % {
                'id': pop.id,
                'add_args': add_args_header,
                'add_args_call': add_args_invoke,
            }
            kernel_decl += CUDATemplates.population_update_kernel['global']['kernel_decl'] % {
                'id': pop.id, 'add_args': add_args_header
            }
            global_call = CUDATemplates.population_update_kernel['global']['host_call'] % {
                'id': pop.id, 'add_args': add_args_call
            }

        # Local variables
        if loc_eqs.strip() != '':
            add_args_header, add_args_invoke, add_args_call = self._gen_kernel_args(pop, 'local')

            # targets
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                add_args_header += """, %(type)s* _sum_%(target)s""" % {'type': ConfigManager().get('precision', self._net_id), 'target' : target}
                add_args_invoke += """, _sum_%(target)s""" % {'target' : target}
                add_args_call += """, pop%(id)s->gpu__sum_%(target)s""" % {'id': pop.id, 'target' : target}

            # global operations
            for op in pop.global_operations:
                ids = {
                    'id': pop.id,
                    'type': ConfigManager().get('precision', self._net_id),
                    'op': op['function'],
                    'var': op['variable']
                }
                add_args_header += """, %(type)s _%(op)s_%(var)s """ % ids
                add_args_invoke += """, _%(op)s_%(var)s """ % ids
                add_args_call += """, pop%(id)s->_%(op)s_%(var)s""" % ids

            # finalize code templates
            device_kernel += CUDATemplates.population_update_kernel['local']['device_kernel'] % {
                'id': pop.id,
                'add_args': add_args_header,
                'pop_size': pop.size,
                'local_eqs': loc_eqs,
                'pre_loop': tabify(pre_loop, 1),
                'post_loop': tabify(rng_post_loop, 1)
            }
            kernel_invoke += CUDATemplates.population_update_kernel['local']['invoke_kernel'] % {
                'id': pop.id,
                'add_args': add_args_header,
                'add_args_call': add_args_invoke,
            }
            kernel_decl += CUDATemplates.population_update_kernel['local']['kernel_decl'] % {
                'id': pop.id,
                'add_args': add_args_header
            }
            local_call = CUDATemplates.population_update_kernel['local']['host_call'] % {
                'id': pop.id,
                'add_args': add_args_call
            }

        # Call statement consists of two parts
        host_call = """
    // Updating the local and global variables of population %(id)s
    if ( pop%(id)s->_active ) {
        %(global_call)s

        %(local_call)s
    }
""" % {'id':pop.id, 'global_call': global_call, 'local_call': local_call}

        if self._prof_gen:
            host_call = self._prof_gen.annotate_update_neuron(pop, host_call)

        return device_kernel, kernel_invoke, kernel_decl, host_call

    def _update_spiking_neuron(self, pop):
        """
        Generate the neural update code for GPU devices. We split up the
        calculation into three parts:

            * evolvement of global differential equations
            * evolvement of local differential equations
            * spike gathering

        Return:

            * tuple of four code snippets (device_kernel, device_invoke, kernel_decl, host_call)
        """
        # Use pre-defined code template
        if 'update_variables' in pop._specific_template.keys():
            try:
                return pop._specific_template['update_variable_body'], pop._specific_template['update_variable_invoke'], pop._specific_template['update_variable_header'], pop._specific_template['update_variable_call']
            except KeyError:
                Messages._error("\nCode generation error: if one attempts to override the population update on CUDA devices, one need to define all of the following fields of _specific_template dictionary:\n\tupdate_variables, update_variable_call, update_variable_header, update_variable_invoke, update_variable_body")

        # Is there any variable?
        if len(pop.neuron_type.description['variables']) == 0:
            return "", "", "", ""

        kernel_decl = ""
        device_kernel = ""
        device_invoke = ""
        local_call = ""
        global_call = ""

        # some defaults
        ids = {
            'id': pop.id,
            'local_index': "[i]",
            'global_index': "[0]"
        }

        #
        # Process the global and local equations and generate first set of kernels
        #

        # parse the equations
        glob_eqs = generate_equation_code(pop.neuron_type.description, locality='global', padding=1)
        loc_eqs = generate_equation_code(pop.neuron_type.description, locality='local', padding=2)

        # Gather pre-loop declaration (dt/tau for ODEs) and
        # update the related kernels
        pre_code = ""
        for var in pop.neuron_type.description['variables']:
            if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                pre_code += ConfigManager().get('precision', self._net_id) + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'
        if pre_code.strip() != '':
            pre_code = """
    // Updating the step sizes
""" + tabify(pre_code, 1)

        # replace pow() for SDK < 6.5
        loc_eqs = check_and_apply_pow_fix(loc_eqs, self._cuda_version)
        glob_eqs = check_and_apply_pow_fix(glob_eqs, self._cuda_version)

        # replace the random distributions
        loc_eqs, glob_eqs, rng_pre_loop, rng_post_loop = self._replace_random(loc_eqs, glob_eqs, pop.neuron_type.description['random_distributions'])
        pre_code += rng_pre_loop

        # within refractory perid, only conductance variables
        if pop.neuron_type.refractory or pop.refractory:
            refr_eqs = generate_equation_code(pop.neuron_type.description, 'local', conductance_only=True, padding=3)
            loc_eqs = tabify(loc_eqs, 1) # just for better code layout
        else:
            refr_eqs = ''

        # Replace %(global_idx)s for global parameters
        for var in pop.neuron_type.description["global"]:
            attr_type, attr_dict = self._get_attr_and_type(pop, var)
            if attr_type == 'par':
                if pre_code.strip() != '':
                    pre_code = pre_code.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if refr_eqs.strip() != '':
                    refr_eqs = refr_eqs.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if glob_eqs.strip() != '':
                    glob_eqs = glob_eqs.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if loc_eqs.strip() != '':
                    loc_eqs = loc_eqs.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])

        # Fill the placeholder
        glob_eqs = glob_eqs % ids
        loc_eqs = loc_eqs % ids
        pre_code = pre_code % ids
        rng_post_loop = rng_post_loop % ids
        refr_eqs = refr_eqs % ids

        # replace local function calls
        if len(pop.neuron_type.description['functions']) > 0:
            glob_eqs, loc_eqs = self._replace_local_funcs(pop, glob_eqs, loc_eqs)

        # Global variables
        if glob_eqs.strip() != '':
            add_args_header, add_args_invoke, add_args_call = self._gen_kernel_args(pop, 'global')

            # global operations
            for op in pop.global_operations:
                ids = {
                    'id': pop.id,
                    'type': ConfigManager().get('precision', self._net_id),
                    'op': op['function'],
                    'var': op['variable']
                }
                add_args_header += """, %(type)s _%(op)s_%(var)s """ % ids
                add_args_invoke += """, _%(op)s_%(var)s """ % ids
                add_args_call += """, pop%(id)s->_%(op)s_%(var)s""" % ids

            # finalize code templates
            body += CUDATemplates.population_update_kernel['global']['body'] % {
                'id': pop.id,
                'add_args': add_args_header,
                'pre_loop': pre_code,
                'post_loop': rng_post_loop,
                'global_eqs':glob_eqs
            }
            header += CUDATemplates.population_update_kernel['global']['header'] % {
                'id': pop.id, 'add_args': add_args_header
            }
            global_call = CUDATemplates.population_update_kernel['global']['call'] % {
                'id': pop.id, 'add_args': add_args_call, 'stream_id': pop.id
            }

        # Local variables
        if loc_eqs.strip() != '':
            add_args_header, add_args_invoke, add_args_call = self._gen_kernel_args(pop, 'local')

            # global operations
            for op in pop.global_operations:
                ids = {
                    'id': pop.id,
                    'type': ConfigManager().get('precision', self._net_id),
                    'op': op['function'],
                    'var': op['variable']
                }
                add_args_header += """, %(type)s _%(op)s_%(var)s """ % ids
                add_args_invoke += """, _%(op)s_%(var)s """ % ids
                add_args_call += """, pop%(id)s->_%(op)s_%(var)s""" % ids

            # Is there a refractory period?
            if pop.neuron_type.refractory or pop.refractory:
                # 'old' loc_eqs is now only executed ouside refractory period
                loc_eqs = """
        // Refractory period
        if( refractory_remaining[i] > 0){
%(refr_eqs)s
            // Decrement the refractory period
            refractory_remaining[i]--;
        } else{
%(loc_eqs)s
        }
""" %  {'refr_eqs': refr_eqs, 'loc_eqs': loc_eqs}

                add_args_header += ", int* refractory_remaining"
                add_args_invoke += ", refractory_remaining"
                add_args_call += """, pop%(id)s->gpu_refractory_remaining""" %{'id':pop.id}

            # finalize code templates
            device_kernel += CUDATemplates.population_update_kernel['local']['device_kernel'] % {
                'id': pop.id, 'add_args': add_args_header, 'pop_size': pop.size, 'pre_loop': tabify(pre_code, 1), 'post_loop': tabify(rng_post_loop, 1), 'local_eqs': loc_eqs
            }
            device_invoke += CUDATemplates.population_update_kernel['local']['invoke_kernel'] % {
                'id': pop.id, 'add_args': add_args_header, 'add_args_call': add_args_invoke
            }
            kernel_decl += CUDATemplates.population_update_kernel['local']['kernel_decl'] % {
                'id': pop.id, 'add_args': add_args_header
            }
            local_call = CUDATemplates.population_update_kernel['local']['host_call'] % {
                'id': pop.id, 'add_args': add_args_call, 'stream_id': pop.id
            }

        # Call statement consists of two parts
        host_call = """
    // Updating the local and global variables of population %(id)s
    if ( pop%(id)s->_active ) {
        %(global_call)s

        %(local_call)s
    }
""" % {'id':pop.id, 'global_call': global_call, 'local_call': local_call}

        if self._prof_gen:
            host_call = self._prof_gen.annotate_update_neuron(pop, host_call)

        return device_kernel, device_invoke, kernel_decl, host_call

    def _spike_gather(self, pop):
        """
        Process the spike condition and generate the corresponding kernel including call code.

        Returns: four separate strings containg:

        *device_kernel*:    device code
        *invoke_kernel*:    defice function invocation
        *kernel_decl*:      header definition for invoke_kernel
        *host_call*:        device function call
        """
        # some defaults
        ids = {
            'id': pop.id,
            'local_index': "[i]",
            'global_index': "[0]"
        }

        if 'spike_gather_body' in pop._specific_template.keys():
            try:
                return pop._specific_template['spike_gather_body'], pop._specific_template['spike_gather_invoke'], pop._specific_template['spike_gather_header'], pop._specific_template['spike_gather_call']
            except KeyError:
                Messages._error("\nCode generation error: if one attempts to override the spike gathering on CUDA devices, one need to define all of the following fields of _specific_template dictionary: spike_gather_call, spike_gather_header, spike_gather_body")

        cond = pop.neuron_type.description['spike']['spike_cond']
        reset = ""
        for eq in pop.neuron_type.description['spike']['spike_reset']:
            reset += """
            %(reset)s
""" % {'reset': eq['cpp']}

        # arguments
        header_args = ""
        header_invoke = ""
        call_args = ""

        # gather all attributes required by this kernel
        kernel_deps = []
        for var in pop.neuron_type.description['spike']['spike_cond_dependencies']:
            kernel_deps.append(var)
        for reset_eq in pop.neuron_type.description['spike']['spike_reset']:
            kernel_deps.append(reset_eq['name'])
            for var in reset_eq['dependencies']:
                kernel_deps.append(var)

        # remove doubled entries and sort to prevent unnecessary recompiles
        kernel_deps = sorted(list(set(kernel_deps)))

        # generate header, call and body args
        for var in kernel_deps:
            attr_type, attr_dict = self._get_attr_and_type(pop, var)

            if attr_type == 'par' and attr_dict['locality'] == "global":
                header_args += ", const " + attr_dict['ctype'] + " " + var
                header_invoke += ", " + var
                call_args += ", pop"+str(pop.id)+"->"+var

                cond = cond.replace(var+"%(global_index)s", var)
                reset = reset.replace(var+"%(global_index)s", var)
            else:
                header_args += ", " + attr_dict['ctype'] + "* " + var
                header_invoke += ", " + var
                call_args += ", pop"+str(pop.id)+"->gpu_"+var

        # Fill the templates with the correct ids
        cond = cond % ids
        reset = reset % ids

        # Is there a refractory period?
        if pop.neuron_type.refractory or pop.refractory:
            # Identify the refractory variable.
            # By default, it is refractory, but users can specify another one
            refrac_var = "refractory[i]"
            if isinstance(pop.neuron_type.refractory, str):
                found = False
                for param in pop.neuron_type.description["parameters"] + pop.neuron_type.description["variables"]:
                    if param["name"] == pop.neuron_type.refractory:
                        if param['locality'] == 'local':
                            refrac_var = "int(" + pop.neuron_type.refractory + "[i]/dt)"
                        else:
                            refrac_var = "int(" + pop.neuron_type.refractory + "/dt)"
                        found = True
                        break
                if not found:
                    Messages._error("refractory = "+ pop.neuron_type.refractory + ": parameter or variable does not exist.")

                refrac_inc = "refractory_remaining[i] = %(refrac_var)s;" % {'refrac_var': refrac_var}
                header_args += ", %(type)s *%(name)s, int* refractory_remaining" % {'type': param['ctype'], 'name': param['name']}
                header_invoke += ", %(name)s, refractory_remaining" % {'name': param['name']}
                call_args += ", pop%(id)s->gpu_%(name)s, pop%(id)s->gpu_refractory_remaining" %{'id':pop.id, 'name': param['name']}

            else: # default case
                refrac_inc = "refractory_remaining[i] = %(refrac_var)s;" % {'refrac_var': refrac_var}
                header_args += ", int *refractory, int* refractory_remaining"
                header_invoke += ", refractory, refractory_remaining"
                call_args += """, pop%(id)s->gpu_refractory, pop%(id)s->gpu_refractory_remaining""" %{'id':pop.id}
        else:
            refrac_inc = ""

        # With ANNarchy 4.7.2 we introduced two different kernel:
        # a) single block (standard version prior to ANNarchy 4.7.2)
        # b) multiple blocks (new in ANNarchy 4.7.2)
        if pop.size < 32:
            launch_config = """int tpb = 32;\nint nb_blocks = 1;\n"""
        else:
            launch_config = """int tpb = 32;\nint nb_blocks = %(nb)s;\n""" % {'nb': int(min(65535, float(pop.size)/32.0))}
        launch_config = tabify(launch_config, 2)

        device_kernel = CUDATemplates.spike_gather_kernel['device_kernel'] % {
            'id': pop.id,
            'pop_size': str(pop.size),
            'float_prec': ConfigManager().get('precision', self._net_id),
            'args': header_args,
            'cond': cond,
            'reset': reset,
            'refrac_inc': refrac_inc
        }

        invoke_kernel = CUDATemplates.spike_gather_kernel['invoke_kernel'] % {
            'id': pop.id,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'args': header_args,
            'args_call': header_invoke
        }

        kernel_decl = CUDATemplates.spike_gather_kernel['kernel_decl'] % {
            'id': pop.id,
            'default': 'const long int t, const %(float_prec)s dt, int* spiked, long int* last_spike' % {'float_prec': ConfigManager().get('precision', self._net_id)},
            'args': header_args
        }

        if pop.max_delay > 1:
            default_args = 't, dt, pop%(id)s->gpu_delayed_spiked.front(), pop%(id)s->gpu_last_spike' % {'id': pop.id}
        else: # no_delay
            default_args = 't, dt, pop%(id)s->gpu_spiked, pop%(id)s->gpu_last_spike' % {'id': pop.id}

        host_call = CUDATemplates.spike_gather_kernel['host_call'] % {
            'id': pop.id,
            'default': default_args,
            'args': call_args % {'id': pop.id},
            'stream_id': pop.id,
            'launch_config': launch_config
        }

        if self._prof_gen:
            host_call = self._prof_gen.annotate_spike_gather(pop, host_call)

        return device_kernel, invoke_kernel, kernel_decl, host_call

    def _memory_transfers(self, pop):
        """
        Before evaluation neuron/synaptic equations we need to update the data
        on the GPU. To synchronize the states of variables after simulation of
        several steps, we need to transfer variables back to the host.

        Return:

            (str, str): host_device_transfer, device_host_transfer

        Notice:

            these codes are part of the run() or step() method (defined in ANNarchy.cu).
        """
        host_device_transfer = ""
        device_host_transfer = ""

        # Variables/Parameters
        host_device_transfer += """
    // host to device transfers for %(name)s""" % {'name': pop.name}
        for attr in pop.neuron_type.description['variables']:
            ids = {'id': pop.id, 'attr_name': attr['name'], 'type': attr['ctype']}
            if attr['name'] in pop.neuron_type.description['local']:
                host_device_transfer += self._templates['attribute_transfer']['HtoD_local'] % ids
            else:
                host_device_transfer += self._templates['attribute_transfer']['HtoD_global'] % ids
        for attr in pop.neuron_type.description['parameters']:
            ids = {'id': pop.id, 'attr_name': attr['name'], 'type': attr['ctype']}
            if attr['name'] in pop.neuron_type.description['local']:
                host_device_transfer += self._templates['attribute_transfer']['HtoD_local'] % ids
            else:
                # nothing to do for global parameter
                continue

        # Rate-coded targets
        if pop.neuron_type.type == "rate":
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                ids = {'attr_name': "_sum_"+target, 'type': ConfigManager().get('precision', self._net_id), 'id': pop.id}
                host_device_transfer += self._templates['attribute_transfer']['HtoD_local'] % ids

        # Refractoriness
        if pop.neuron_type.type == "spike":
            if pop.neuron_type.refractory or pop.refractory:
                host_device_transfer += """
        // refractory
        if( refractory_dirty )
        {
        #ifdef _DEBUG
            std::cout << "HtoD refractory ( pop%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_refractory, refractory.data(), size * sizeof(int), cudaMemcpyHostToDevice);
            refractory_dirty = false;

        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""" % {'id': pop.id}

        # Write back variables. The operation should be performed only if it hasn't done
        # since the last simulate() / step() and for a given variable. 
        for attr in pop.neuron_type.description['variables']:
            ids = {'attr_name': attr['name'], 'type': attr['ctype'], 'id': pop.id}
            if attr['name'] in pop.neuron_type.description['local']:
                device_host_transfer += self._templates['attribute_transfer']['DtoH_local'] % ids
            else:
                device_host_transfer += self._templates['attribute_transfer']['DtoH_global'] % ids

        # Write back rate-coded targets
        if pop.neuron_type.type == "rate":
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                ids = {'attr_name': "_sum_"+target, 'type': ConfigManager().get('precision', self._net_id), 'id': pop.id}
                device_host_transfer += self._templates['attribute_transfer']['DtoH_local'] % ids

        if 'host_device_transfer' in pop._specific_template.keys():
            host_device_transfer = pop._specific_template['host_device_transfer']
        if 'device_host_transfer' in pop._specific_template.keys():
            device_host_transfer = pop._specific_template['device_host_transfer']

        return host_device_transfer, device_host_transfer
